package imp.ringing;

import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ArrayBlockingQueue;

import android.util.Log;

/**
 * A protocol that encodes data into alternating blocks of ringing and silence.
 */
public class AlternatingProtocol extends Protocol {
	
	private final long GRANULARITY_MS;
	private final int BLOCK_SIZE_R;
	private final int BLOCK_SIZE_S;
	private final ArrayBlockingQueue<Boolean> queue;
	private long elapsedTime;
	private long ringingElapsedTime;
	private long silentElapsedTime;
	private boolean ringingPhase = false;
	
	/**
	 * Computes the bit transmission time for given block sizes.
	 * @param r Ringing block size.
	 * @param s Silent block size.
	 * @return Bit transmission time for block sizes r, s.
	 */
	private double getTransmissionTime(int r, int s) {
		return (params.getDMax() + params.getHMax() + 
				GRANULARITY_MS * (Math.pow(2, r) - 1) / 2.0 +
				GRANULARITY_MS * (Math.pow(2, s) - 1) / 2.0
				) / (double)(r + s);
	}
	
	/**
	 * Initialize the protocol and compute the optimal block sizes.
	 * @param params Parameters.
	 */
	public AlternatingProtocol(Parametrization params) {
		super(params);
		
		GRANULARITY_MS = params.getDeltaD() + params.getDeltaH();
		long limit = params.getRingLength() - params.getHMax() - params.getDeltaD();
		
		// Find the optimal block size of the form [b, b].
		int bestBlockSize = 0;
		double bestBlockSizeTime = 0; 
		
		for (int i = 1; ; i++) {
			double time = getTransmissionTime(i, i);
			if (bestBlockSize == 0 || time < bestBlockSizeTime) {
				bestBlockSize = i;
				bestBlockSizeTime = time;
			} else break;
		}
		
		int bestBlockSizeR = bestBlockSize;
		int bestBlockSizeS = bestBlockSize;
		
		// If the optimal value is feasible, use it or one of its neighbors.
		if (GRANULARITY_MS * (Math.pow(2, bestBlockSizeR) - 1) <= limit) {
				
			// Try the neighbors.
			if (bestBlockSize > 1 && getTransmissionTime(bestBlockSize - 1, bestBlockSize) < bestBlockSizeTime) {
				bestBlockSizeR = bestBlockSizeR - 1;
				bestBlockSizeS = bestBlockSize;
				bestBlockSizeTime = getTransmissionTime(bestBlockSize - 1, bestBlockSize);
			}
			if (getTransmissionTime(bestBlockSize, bestBlockSize + 1) < bestBlockSizeTime) {
				bestBlockSizeR = bestBlockSize;
				bestBlockSizeS = bestBlockSizeS + 1;
				bestBlockSizeTime = getTransmissionTime(bestBlockSize, bestBlockSize + 1);
			}

		} else { // If ringing block size is too large, find the best silent block size.
			
			// Decrease ringing block size to the maximum feasible value.
			while (GRANULARITY_MS * (Math.pow(2, bestBlockSizeR) - 1) > limit && bestBlockSizeR > 1) {
				--bestBlockSizeR;
			}
			
			// Find the optimal silent block size.
			bestBlockSizeS = 0;
			bestBlockSizeTime = 0; 			
			
			for (int i = 1; ; i++) {
				double time = getTransmissionTime(bestBlockSizeR, i); 
				if (bestBlockSizeS == 0 || time < bestBlockSizeTime) {
					bestBlockSizeS = i;
					bestBlockSizeTime = time;
				} else break;
			}
			
		}
		
		BLOCK_SIZE_R = bestBlockSizeR;
		BLOCK_SIZE_S = bestBlockSizeS;
				
		queue = new ArrayBlockingQueue<Boolean>(Math.max(BLOCK_SIZE_R, BLOCK_SIZE_S));
		Log.i("USER", "Block sizes: R=" + BLOCK_SIZE_R + ", S=" + BLOCK_SIZE_S);		
	}

	@Override
	public void send(boolean b, boolean last) throws InterruptedException {
		Log.i("USER", "Sender: Queuing bit " + (b ? "1" : "0"));
		queue.offer(b);
		
		// Do not send, only add bit to the queue.
		if (queue.size() < (ringingPhase ? BLOCK_SIZE_S : BLOCK_SIZE_R) && !last) {
			return;
		}
		
		// Pad the end of the last block.
		while (queue.size() < (ringingPhase ? BLOCK_SIZE_S : BLOCK_SIZE_R)) {
			queue.offer(false);
		}
	
		// If the queue is full, construct a value from its bits.
		elapsedTime = 0;
		for (int i = 0; i < (ringingPhase ? BLOCK_SIZE_S : BLOCK_SIZE_R); i++)
			if (queue.poll())
				elapsedTime |= 1 << i;

		elapsedTime *= GRANULARITY_MS;
		Log.i("USER", "Sender: Expected block time " + elapsedTime);

		// Send.
		queue.clear();
		
		ringingPhase = !ringingPhase;
		
		if (ringingPhase) {
			// Ring...
			ring();
			// ...until time hangup.
			this.mutexUp();
			// Allow h_{max} time for hangup.
			Thread.sleep(params.getHMax());
		} else {
			silentElapsedTime = this.getElapsedTimeMillis();
			// Wait for the computed amount of time.
			Thread.sleep(elapsedTime);
			Log.i("USER", "Sender: Quiet ended after " + (this.getElapsedTimeMillis() - silentElapsedTime));
		}

		// Always end with a ringing block.
		if (last && !ringingPhase) {
			for (int i = 0; i < BLOCK_SIZE_R; i++)
				send(false, false);
		}
	}

	@Override
	public void callingCallback() {
		Timer timer = new Timer();
		
		// Allow d_{max} time for dialing, then ring for the computed amount of time.
		timer.schedule(new TimerTask() {
			@Override
			public void run() {
				hangup();
				Log.i("USER", "Sender: Hangup");
			}
		}, elapsedTime + params.getDMax());
		
		ringingElapsedTime = this.getElapsedTimeMillis();
	}

	@Override
	public void callingHangupCallback() {
		Log.i("USER", "Sender: Call ended after " + (this.getElapsedTimeMillis() - ringingElapsedTime));
		this.mutexDown();
	}

	@Override
	public void receivingCallback(String number) {
		// End silence timing, start ringing timing. 
		ringingElapsedTime = getElapsedTimeMillis();
		silentElapsedTime = getElapsedTimeMillis() - silentElapsedTime;		
		if (!ringingPhase) {
			this.mutexDown();
		}
	}

	@Override
	public void receivingHangupCallback() {
		// End ringing timing, start silence timing.
		silentElapsedTime = getElapsedTimeMillis();
		ringingElapsedTime = getElapsedTimeMillis() - ringingElapsedTime;		
		if (ringingPhase) {
			this.mutexDown();
		}
	}
	
	@Override
	public boolean receive() throws InterruptedException {
		if (queue.isEmpty()) {
			// Wait for the next event.
			ringingPhase = !ringingPhase;
			Log.i("USER", "Recvr : Waiting for " + (ringingPhase ? "ringing" : "silence") + " to end");
			this.mutexUp();
			
			// Decode.
			long elapsedTime = ringingPhase ? ringingElapsedTime : silentElapsedTime;
			Log.i("USER", "Recvr : Elapsed time: " + elapsedTime + " ms");
			int number = (int)((elapsedTime - (ringingPhase ? params.getHMin() : params.getDMin())) / GRANULARITY_MS);
			Log.i("USER", "Recvr : Received as " + number);
			
			// Deconstruct the value to bits.
			queue.clear();
			for (int i = 0; i < (ringingPhase ? BLOCK_SIZE_R : BLOCK_SIZE_S); i++)
				queue.offer((number & (1 << i)) != 0);						
		}

		Log.i("USER", "Recvr : Received bit " + (queue.peek() ? "1" : "0"));
		return queue.poll();
	}

}
